/* $USAGI: ipsec6_main.c,v 1.22 2001/09/21 09:25:13 mk Exp $ */

/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <asm/byteorder.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sysctl.h>
#include <linux/inet.h>
#include <linux/ipv6.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/smp.h>
#include <linux/list.h>
#include <asm/uaccess.h>
#include <net/ipv6.h>
#include <net/sadb.h>
#include <net/spd.h>
#include <net/ip6_route.h>
#include <net/addrconf.h>
#include <net/snmp.h>  
#include <net/ipsec6_utils.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/icmpv6.h>
#include <linux/ipsec.h>
#include <linux/ipsec6.h>
#include <net/pfkeyv2.h> /* sa proto type */
#include <net/pfkey.h>

#include <linux/ip6packettools.h> /* XXX: analyze_ip6_packet */

#include "calc_hmac.h"

/* DEBUG *******************************************/
#ifdef CONFIG_IPV6_IPSEC_DEBUG
#define ipsec6_pr(fmt, arg...) printk(KERN_DEBUG fmt,##arg)
#else
#define ipsec6_pr(fmt,arg...) \
        do { } while (0)
#endif /* CONFIG_IPV6_IPSEC_DEBUG */

/**************************************************/

#ifdef CONFIG_SYSCTL

#if 0

/* XXX insert linux/sysctl.h */
enum {
        NET_IPSEC_CONF_IPV6  = 1,
};

static struct ctl_table_header *ipsec_table_header;

/* XXX please extern anywhare *.h !!!*/
int sysctl_ipsec6 = 1;

static ctl_table ipsec_table[] = {
	{ NET_IPSEC_CONF_IPV6,
	  "ipsec6",
	  &sysctl_ipsec6,
	  sizeof(int),
	  0644,
	  NULL,
	  proc_dointvec,
	  &sysctl_intvec,
	  NULL, NULL, NULL
	},
	{0} 
};

/* XXX insert NET_IPEC enum in CTL_NET  in linux/sysctl.h */
static ctl_table ipsec_dir_table[] = {
        {NET_IPSEC, "ipsec", NULL, 0, 0555, ipsec_table,0,0,0,0,0},
        {0}
};

static ctl_table ipsec_root_table[] = {
        {CTL_NET, "net", NULL, 0, 0555, ipsec_dir_table,0,0,0,0,0},
        {0}
};
        
void init_ipsec_sysctl(void)
{
	ipsec_table_header = register_sysctl_table(ipsec_root_table, 1);
}

void cleanup_ipsec_sysctl(void)
{
	unregister_sysctl_table(ipsec_table_header);
}
#endif /* 0 */
#endif /* CONFIG_SYSCTL */
/**********************************************************************************/
int ipsec6_init(void)
{
	int result;
#if 0	
	result=algos_init();
	if (result!=0) {
	        return result;
	}
#endif
	result=pfkey_init();
	if (result!=0) {
//		algos_cleanup();
		return result;
	}

#if 0
	result = ksadbd_init();
	if (result != 0) {
		ksadbd_exit();
		return result;
	}
#endif
	printk(KERN_INFO "IPv6/IPsec registered\n");

	return 0;
}

void ipsec6_cleanup (void)
{
	pfkey_cleanup();
//	algos_cleanup();

	printk(KERN_INFO "IPv6/IPsec unregistered\n");
}

/**********************************************************************************/
#if 0
enum {
	    ReplayWindowSize = 32
};
#endif /* 0 */

/* XXX must insert window size  sa_replay_window member */
int ChkReplayWindow(struct sa_replay_window *rw, __u32 hdrseq)
{       
        __u32 diff;
	__u32 seq = ntohl(hdrseq);

	ipsec6_pr(
"ChkReplayWindow: overflow: %u
                      size: %u
                   seq_num: %x
                  last_seq: %x
                    bitmap: %08x
                  curr seq: %x\n",
			rw->overflow, rw->size, rw->seq_num, rw->last_seq, rw->bitmap, seq);
	if (seq == 0) {
		return 0; /* first == 0 or wrapped */
	}

	if (seq > rw->last_seq) { /* new larger sequence number */
		diff = seq - rw->last_seq;
		if (diff < rw->size) {  /* In window */
			rw->bitmap <<= diff;
			rw->bitmap |= 1; /* set bit for this packet */
		} else {
			rw->bitmap = 1; /* This packet has a "way larger" */
		}

		rw->last_seq = seq;
		return 1; /* larger is good */
	}

	diff = rw->last_seq - seq;

	if (diff >= rw->size) {
	       return 0; /* too old or wrapped */
	}

	if ( rw->bitmap & ((u_long)1 << diff) ) {
	       return 0; /* already seen */
	}

	rw->bitmap |= ((u_long)1 << diff); /* mark as seen */

	return 1; /* out of order but good */
}

int ipsec6_input_check(struct sk_buff **skb, int *nexthdr)
{
	int len;
	int rtn = 0;
	struct ip6_packet_refs refs; /* XXX */
	struct ipsec_sa *ipsec_sa = NULL;
	struct sa_index auth_sa_idx;
	struct sa_index esp_sa_idx;
	struct selector selector;
	struct ipsec_sp *policy = NULL;
	int error;
	int addr_type = 0;
	struct ipv6hdr *hdr = (*skb)->nh.ipv6h;

	ipsec6_pr("ipsec6_input_check: called\n");
#ifdef CONFIG_IPV6_IPSEC_DEBUG
	{
		char buf[64];
		ipsec6_pr("ipsec6_input_check: dst addr: %s\n", 
			  in6_ntop( &hdr->daddr, buf));
		ipsec6_pr("ipsec6_input_check: src addr: %s\n", 
			  in6_ntop( &hdr->saddr, buf));
		ipsec6_pr("ipsec6_input_check: hdr->payload_len is %d\n", ntohs(hdr->payload_len)); /* XXX: remove before commit (mk) */
	}
#endif /* CONFIG_IPV6_IPSEC_DEBUG */
	/* XXX */
	addr_type = ipv6_addr_type(&hdr->daddr);
	if (addr_type & IPV6_ADDR_MULTICAST) {
		ipsec6_pr("ipsec6_input_check: address type multicast skip!\n");
		goto finish;
	}
	
	if (!skb || !*skb) {
		printk(KERN_ERR "ipsec6_input_check: skb NULL!\n");
		rtn =  -EINVAL;
		goto finish;
	}

	if ( !((*skb)->nh.ipv6h) ) {
		ipsec6_pr("ipsec6_input_check: nh.ipv6h NULL!\n");
		rtn = -EINVAL;
		goto finish;
	} else if ( (*skb)->nh.ipv6h->nexthdr == NEXTHDR_UDP ) { /* IKE */
		(*skb)->h.uh = (void *)((char*)((*skb)->nh.ipv6h) + sizeof(struct ipv6hdr));
		if ( (*skb)->h.uh != NULL && (ntohs((*skb)->h.uh->source) == 500) ) {
			ipsec6_pr("received IKE packet. skip!\n");
			rtn = 0; /* through */ 
			goto finish;
		}
	} else if ( (*skb)->nh.ipv6h->nexthdr == NEXTHDR_ICMP ) { 
		ipsec6_pr("ipsec6_input_check: received ICMP packet.\n");
#ifdef CONFIG_APPLY_ICMPV6_IPSEC /* XXX currently we can't handle NS-NA -mk */
		(*skb)->h.icmp6h = (void *)((char*)((*skb)->nh.ipv6h) + sizeof(struct ipv6hdr));
		if ( (*skb)->h.icmp6h != NULL) {
			struct icmp6hdr *icmph = (struct icmp6hdr *)((*skb)->h.icmp6h);

			if (icmph->icmp6_type == NDISC_NEIGHBOUR_ADVERTISEMENT) {
				ipsec6_pr("ipsec6_input_check: icmp6 type NA, skip!\n");
				rtn = 0;
				goto finish;
			} else if (icmph->icmp6_type == NDISC_ROUTER_ADVERTISEMENT) {
				ipsec6_pr("ipsec6_input_check: icmp6 type RA, skip!\n");
				rtn = 0;
				goto finish;
			} else {
				ipsec6_pr("ipsec6_input_check: icmp6_type: %u\n", icmph->icmp6_type);
			}
		}
#else
		ipsec6_pr("ipsec6_input_check: skip ICMP packet.\n");
		rtn = 0; 
		goto finish; 
#endif /* CONFIG_APPLY_ICMPV6_IPSEC */
	}

	len = ntohs((*skb)->nh.ipv6h->payload_len) + sizeof(struct ipv6hdr);
	analyze_ip6_packet((char *)(*skb)->nh.ipv6h, &refs);


	/* XXX handle SA bundle */
	sa_index_init(&auth_sa_idx);
	sa_index_init(&esp_sa_idx);

	/* copy selector XXX port */
	memset(&selector, 0, sizeof(struct selector));
	((struct sockaddr_in6 *)&(selector.dst))->sin6_family = AF_INET6;
	ipv6_addr_copy(&((struct sockaddr_in6 *)(&selector.src))->sin6_addr,
		       &((*skb)->nh.ipv6h->saddr));
	((struct sockaddr_in6 *)&(selector.src))->sin6_family = AF_INET6;
	ipv6_addr_copy(&((struct sockaddr_in6 *)(&selector.dst))->sin6_addr,
		       &((*skb)->nh.ipv6h->daddr));
	selector.prefixlen_d = 128;
	selector.prefixlen_s = 128;

	/* AH */
	if (refs.authhdr) {
		__u8* authdata;
		int err;
		size_t authsize;
		char *packet;
		struct ip6_packet_refs packetrefs;

		ipsec6_pr("ipsec6_input_check: start auth header processing\n");
		/* Check SPI */
		ipsec6_pr("ipsec6_input_check: refs.authhdr->spi is 0x%x\n", ntohl(refs.authhdr->spi));

		err = sadb_find_by_address_proto_spi((struct sockaddr *)&selector.src, selector.prefixlen_s,
						     (struct sockaddr *)&selector.dst, selector.prefixlen_d,
						     NULL, 0,			/* proxy address */
						     SADB_SATYPE_AH,
						     refs.authhdr->spi,
						     &ipsec_sa);

		if (err != -EEXIST) {
			printk(KERN_ERR "ipsec6_input_check: not found SPI for ah\n");
			rtn = -EINVAL;
			goto finish;
		}
		if (!ipsec_sa) {
			printk(KERN_ERR "ipsec6_input_check: not found SA for ah\n");
			rtn = -EINVAL;
			goto finish;
		}
		write_lock(&(ipsec_sa->lock));

		switch(ipsec_sa->state) {

		case SADB_SASTATE_LARVAL:
			ipsec6_pr("ipsec6_input_check: sa is larval\n");
			;
			break;

		case SADB_SASTATE_MATURE:
			ipsec6_pr("ipsec6_input_check: sa is mature\n");
			;
			break;


		case SADB_SASTATE_DYING:
			ipsec6_pr("ipsec6_input_check: sa is dying\n");
			/* It needs update. */
			;	
			break;

		case SADB_SASTATE_DEAD:
			ipsec6_pr("ipsec6_input_check: sa is dead\n");
			write_unlock(&(ipsec_sa->lock));
			sadb_remove(ipsec_sa);
			rtn = -ESRCH;
			goto finish;
			;
			break;

		}

		ipsec6_pr("ipsec6_input_check: dump ipsec_sa->spi=0x%x\n", 
				ntohl(ipsec_sa->spi));
		ipsec6_pr("ipsec6_input_check: dump ipsec_sa->ipsec_proto=%u\n",
				ipsec_sa->ipsec_proto);

		if (ipsec_sa->auth_algo.algo == SADB_AALG_NONE ) {
			printk(KERN_ERR "ipsec_input_check: not found auth algo.\n");
			rtn = -EINVAL;
			goto unlock_finish;
		}

		if (!ChkReplayWindow(&(ipsec_sa->replay_window), refs.authhdr->seq_no)) {
		        printk(KERN_ERR "ipsec6_input_check: replay check err!\n");
		        rtn = -EINVAL;
		        goto unlock_finish;
		}

		authsize = ((char *)(*skb)->tail) - ((char *)(*skb)->nh.ipv6h);
		packet= kmalloc(authsize, GFP_ATOMIC); /* ATOMIC ok ? */
		authdata = kmalloc(ipsec_sa->auth_algo.digest_len, GFP_ATOMIC);
		if (!packet || !authdata ) {
		        printk(KERN_ERR "ipsec6_input_check: can't get memory for pakcet\n");
		        rtn = -ENOMEM;
		        goto unlock_finish;
		}

		memcpy(packet, (char*)((*skb)->nh.ipv6h), authsize);
		analyze_ip6_packet(packet, &packetrefs);
		prepare_hash_in_calculation(&packetrefs);
		calc_hmac(ipsec_sa, packet, authsize, authdata);         

		/* Originally, IABG uses "for" loop for matching authentication data.	*/
		/* I change it into memcmp routine.					*/
		if (memcmp(authdata, refs.authhdr->auth_data, ipsec_sa->auth_algo.digest_len)) {
			printk(KERN_ERR "ipsec6_input_check: invalid checksum in AH.\n");
			kfree(packet);
			kfree(authdata);
			rtn = -EINVAL;
			goto unlock_finish;
		}
		kfree(packet);
		kfree(authdata);
		ipsec6_pr("ipsec6_input_check: dump ipsec_sa->spi=0x%x\n",
				ipsec_sa->spi);

		ipsec6_pr("ipsec6_input_check: dump ipsec_sa->ipsec_proto=%u\n",
				ipsec_sa->ipsec_proto);

		error = sa_index_copy_from_sa(&auth_sa_idx, ipsec_sa); 
		if (error < 0) {
			rtn = -EINVAL;
			goto unlock_finish;
		}

		write_unlock(&(ipsec_sa->lock));  /* unlock SA */

		sadb_release_ref(ipsec_sa);

		(*skb)->security |= RCV_AUTH; /* ? we must rewrite linux/ipsec.h */
	}

	if (refs.resttype == NEXTHDR_ESP) {
		int err;
		struct ipv6_esp_hdr *esphdr = (struct ipv6_esp_hdr *)refs.rest;
		u8 *authdata;
		int totalsize, hashsize, encsize;
		int hash_ok;
		u8 *srcdata;
		int srcsize;

		ipsec6_pr("ipsec6_input_check: start esp processing\n");
		/* Check SPI */
		ipsec6_pr("ipsec6_input_check: esphdr->spi is 0x%x\n", ntohl(esphdr->spi));

		err = sadb_find_by_address_proto_spi((struct sockaddr *)&selector.src, selector.prefixlen_s,
						     (struct sockaddr *)&selector.dst, selector.prefixlen_d,
						     NULL, 0,		/* proxy address */
						     SADB_SATYPE_AH,
						     refs.authhdr->spi,
						     &ipsec_sa);

		if (err != -EEXIST) {
			printk(KERN_ERR "ipsec6_input_check: not found SPI for esp\n");
			rtn = -EINVAL;
			goto finish;
		}
		if (!ipsec_sa) {
			printk(KERN_ERR "ipsec6_input_check: not found SA for esp\n");
			rtn = -EINVAL;
			goto finish;
		}
		write_lock(&(ipsec_sa->lock));

		switch(ipsec_sa->state) {

		case SADB_SASTATE_LARVAL:
			ipsec6_pr("ipsec6_input_check: sa is larval\n");
			;
			break;

		case SADB_SASTATE_MATURE:
			ipsec6_pr("ipsec6_input_check: sa is mature\n");
			;
			break;

		case SADB_SASTATE_DYING:
			ipsec6_pr("ipsec6_input_check: sa is dying\n");
			/* It needs update. */
			;	
			break;

		case SADB_SASTATE_DEAD:
			ipsec6_pr("ipsec6_input_check: sa is dead\n");
			write_unlock(&(ipsec_sa->lock));
			sadb_remove(ipsec_sa);
			rtn = -ESRCH;
			goto finish;
			;
			break;

		}

		ipsec6_pr("ipsec6_input_check: dump ipsec_sa->spi=0x%x\n",
				ntohl(ipsec_sa->spi));
		ipsec6_pr("ipsec6_input_check: dump ipsec_sa->ipsec_proto=%u\n",
				ipsec_sa->ipsec_proto);

		ipsec6_pr("ipsec_input_check: use kerneli version.\n");
		if ( ipsec_sa->esp_algo.algo == SADB_EALG_NONE ) {
			printk(KERN_ERR "ipsec6_input_check: not found encryption algorithm in SA!\n");
			rtn = -EINVAL;
			goto unlock_finish;
		}
		totalsize = len - (refs.rest - ((char *)(*skb)->nh.ipv6h));
		encsize = totalsize - ipsec_sa->esp_algo.cx.ci->ivsize - 8;
							/* 8 = SPI + Sequence Number */	
		if ( ipsec_sa->auth_algo.algo != SADB_AALG_NONE ) {
			/* Calculate size */
			/* The tail of payload does not have to be aligned		*/
			/* with a multiple number of 64 bit.				*/
			/* 64 bit alignment is adapted to the position of top of header.*/
			hashsize = ipsec_sa->auth_algo.digest_len;
			encsize -= hashsize;
			authdata=kmalloc(ipsec_sa->auth_algo.digest_len, GFP_ATOMIC);
			calc_hmac(ipsec_sa, (char*)esphdr, totalsize - hashsize, authdata);          
			hash_ok = 1;

			/* Originally, IABG uses "for" loop for matching authentication data. */
			/* I change it into memcmp routine. */

			if (memcmp(authdata, &(((char*)esphdr)[totalsize - hashsize]),
					ipsec_sa->auth_algo.digest_len )) {
				if (net_ratelimit())
					printk(KERN_ERR "ipsec6_input_check: invalid checksum in ESP\n");
				rtn = -EINVAL;
				goto unlock_finish;
			}
			kfree(authdata);
			authdata = NULL;
		}

		/* Decrypt data */
		srcdata = kmalloc(encsize, GFP_ATOMIC);
		if (!srcdata) {
			if (net_ratelimit())
				printk(KERN_ERR "ipsec6_input_check: can't allocate memory for decrypt\n");
			rtn = -ENOMEM;
			goto unlock_finish;
		}

		ipsec6_pr("ipsec6_input_check: len=%d, totalsize=%d, encsize=%d\n",
				len, totalsize, encsize);
		memset(ipsec_sa->esp_algo.cx.iv, 0, MAX_IV_SIZE);
		memcpy(ipsec_sa->esp_algo.cx.iv, esphdr->enc_data, ipsec_sa->esp_algo.cx.ci->ivsize);
		ipsec_sa->esp_algo.cx.ci->decrypt_atomic(&(ipsec_sa->esp_algo.cx),
				((char*)(esphdr->enc_data)) + ipsec_sa->esp_algo.cx.ci->ivsize,
				srcdata, encsize);

		ipsec6_pr("ipsec6_input_check(): Original data was\n");
				
		/* encsize - (pad_len + next_hdr) - pad_len */
		srcsize = encsize - 2 - srcdata[encsize-2];
		ipsec6_pr("ipsec6_input_check: srcsize=%d, padlength=%d\n",
				srcsize, srcdata[encsize-2]);
		if (srcsize <= 0) {
			if (net_ratelimit())
				printk(KERN_ERR "ipsec6_input_check: Encrypted packet contains garbage(Size of decrypted packet < 0).\n");
			rtn = -EINVAL;
			goto unlock_finish;
		}

		if ( !ChkReplayWindow(&(ipsec_sa->replay_window), esphdr->seq_no) ) {
			printk(KERN_ERR "ipsec6_input_check: replay check err!\n");
			rtn = -EINVAL;
			goto unlock_finish;
		}
		*nexthdr = srcdata[encsize-1];
		refs.resttype = srcdata[encsize-1];
		memcpy(refs.rest, srcdata, srcsize);

		skb_trim(*skb, (*skb)->len +  srcsize - totalsize);
		(*skb)->nh.ipv6h->payload_len = htons(((char *)esphdr - (char *)((*skb)->nh.ipv6h))  - sizeof(struct ipv6hdr) + srcsize); /* ok ? -mk */
		if (refs.ipv6hdr->nexthdr == NEXTHDR_ESP) { 
			refs.ipv6hdr->nexthdr = *nexthdr;
		} else if (refs.authhdr && refs.authhdr->nexthdr == NEXTHDR_ESP) {
			refs.authhdr->nexthdr = *nexthdr;
		} else if (refs.destopthdr && refs.destopthdr->nexthdr == NEXTHDR_ESP) {
			refs.destopthdr->nexthdr = *nexthdr;
		} else if (refs.hopopthdr && refs.hopopthdr->nexthdr == NEXTHDR_ESP) {
			refs.hopopthdr->nexthdr = *nexthdr;
		}

		kfree(srcdata);
		srcdata = NULL;

		error = sa_index_copy_from_sa(&esp_sa_idx, ipsec_sa);
		if (error < 0) {
			rtn = -EINVAL;
			goto unlock_finish;
		}

		/* Otherwise checksum of fragmented udp packets fails (udp.c, csum_fold) */
		(*skb)->ip_summed = CHECKSUM_UNNECESSARY; 
		(*skb)->security |= RCV_CRYPT;
		write_unlock(&(ipsec_sa->lock)); /* unlock SA */

		sadb_release_ref(ipsec_sa);
		
	}

	if (refs.resttype == NEXTHDR_ICMP) {
		ipsec6_pr("ipsec6_input_check: nexthdr: icmp.\n");
		selector.proto = IPPROTO_ICMPV6;
	}

	if (refs.resttype == NEXTHDR_TCP) {
		ipsec6_pr("ipsec6_input_check: nexthdr: tcp.\n");
		selector.proto = IPPROTO_TCP;
	}

	if (refs.resttype == NEXTHDR_UDP) {
		ipsec6_pr("ipsec6_input_check: nexthdr: udp.\n");
		selector.proto = IPPROTO_UDP;
	}

	/* beggining of matching check selector and policy */
	ipsec6_pr("ipsec6_input_check: start match check SA and policy.\n");

#ifdef CONFIG_IPV6_IPSEC_DEBUG
	{
		char buf[64];
		ipsec6_pr("ipsec6_input_check: selector dst addr: %s\n", 
			  in6_ntop( &((struct sockaddr_in6 *)&(selector.dst))->sin6_addr, buf));
		ipsec6_pr("ipsec6_input_check: selector src addr: %s\n", 
			  in6_ntop( &((struct sockaddr_in6 *)&(selector.src))->sin6_addr, buf));
	}
#endif /* CONFIG_IPV6_IPSEC_DEBUG */
	error = spd_find_by_selector(&selector, &policy, SPD_WITH_RLOCK);
		
	if (error == -EEXIST) {
		struct list_head *pos = NULL;
		struct sa_index *tmp = NULL;
			
		list_for_each(pos, &(policy->sa_index_list)) {
			tmp = list_entry(pos, struct sa_index, entry);
			if ( tmp->ipsec_proto == SADB_SATYPE_AH &&
				(sa_index_compare(&auth_sa_idx, tmp)) != 0 ) {
				/* not match */
				printk(KERN_ERR "ipsec6_input_check: not match policy and auth SA.\n");
				rtn = -EINVAL;
				break;
			}
			if( tmp->ipsec_proto == SADB_SATYPE_ESP && 
				(sa_index_compare(&esp_sa_idx, tmp)) !=0 ){
				/* not match */
				printk(KERN_ERR "ipsec6_input_check: not match policy and esp SA.\n");
				rtn = -EINVAL;
				break;
			}
		}
		SPD_RUNLOCK(&(policy->lock));
	} else if (error == -ESRCH) { /* XXX ask miyazawa ! mk 28/05/2001 */ 
		ipsec6_pr("ipsec6_input_check:  matching pair of SA and policy not found, through!\n"); 
		goto finish;		
	} else {
		rtn = error;
	}
	ipsec6_pr("ipsec6_input_check: end match check SA and policy.\n");
	/* end of matching check selector and policy */
		
	goto finish;
 unlock_finish:
	write_unlock(&(ipsec_sa->lock));		
 finish:
	return rtn;
}

static int ipsec6_output_check_core(struct selector *selector, void **ipsec_ptr)
{
	int error = 0;
	int renew_sa_list = 0;
	struct ipsec_sp *policy = NULL;
	struct sa_index *sa_idx = NULL;
	struct list_head *pos = NULL; 		/* using at list_for_each */
	struct list_head *sa_index_list = NULL;
	int result = IPSEC_ACTION_BYPASS; 	/* default */

	ipsec6_pr("ipsec6_output_check_core: called\n");

	if (!selector) {
		ipsec6_pr("ipsec6_output_check_core: selector is NULL\n");
		error = -EINVAL;
		goto err;
	}

	error = spd_find_by_selector(selector, &policy, SPD_WITH_RLOCK);
	if (error != -EEXIST) { /* not match ! */
		ipsec6_pr("ipsec6_output_check_core: policy no exists(spd_find_by_selector).\n");
		result = IPSEC_ACTION_BYPASS;
		goto err;
	}

	sa_index_list = kmalloc(sizeof(struct list_head), GFP_KERNEL);
	if (!sa_index_list) {
		ipsec6_pr("ipsec6_output_check_core: kmalloc falid\n");
		error = -ENOMEM;
		goto err;
	}

	INIT_LIST_HEAD(sa_index_list);
	sa_list_copy(sa_index_list, &(policy->sa_index_list));
	SPD_RUNLOCK(&(policy->lock));

	/* sadb lookup */
	/* XXX TODO: If sa_idx->sa is null, SADB_ACQUIRE message should be thrown up to KMd. */
	/* renew_sa_list is use as a flag which indicates
		policy's sa_list was renew by sadb_find_by_sa_index. */
	list_for_each(pos, sa_index_list) {
		sa_idx = list_entry(pos, struct sa_index, entry);

		if (sa_idx) {
			if (!(sa_idx->sa)) { 
				error = sadb_find_by_sa_index(sa_idx);
				if (error != -EEXIST) {
					printk(KERN_ERR "ipsec6_output_check_core: can't find SA\n");
					result = IPSEC_ACTION_DROP;
					goto err;
				} else {
					renew_sa_list = 1;
					printk(KERN_DEBUG "ipsec6_output_check_core: found SA\n");
				}
			}

			write_lock(&(sa_idx->sa->lock));
			switch(sa_idx->sa->state) {

				case SADB_SASTATE_LARVAL:
					ipsec6_pr("ipsec6_input_check: sa is larval\n");
					write_unlock(&(sa_idx->sa->lock));
					sa_list_clear(sa_index_list);
					kfree(sa_index_list);
					result = IPSEC_ACTION_DROP;
					goto err;

				case SADB_SASTATE_MATURE:
					ipsec6_pr("ipsec6_input_check: sa is mature\n");
					;
					break;

				case SADB_SASTATE_DYING:
					ipsec6_pr("ipsec6_input_check: sa is dying\n");
					/* It needs update. */
					;	
					break;

				case SADB_SASTATE_DEAD:
					ipsec6_pr("ipsec6_input_check: sa is dead\n");
					write_unlock(&(sa_idx->sa->lock));
					sadb_remove(sa_idx->sa);
					result = IPSEC_ACTION_DROP;
					goto err;
			}

			switch (sa_idx->sa->ipsec_proto) { /* SADB_* defined in pfkeyv2.h */
				case SADB_SATYPE_AH: 
					result |= IPSEC_ACTION_AUTH;
					break;
				case SADB_SATYPE_ESP:
					result |= IPSEC_ACTION_ESP;
					break;
				default:
					result = IPSEC_ACTION_DROP;
					write_unlock(&(sa_idx->sa->lock));
					goto err;
			}
			write_unlock(&(sa_idx->sa->lock));

		} else {
			printk(KERN_ERR "ipsec6_output_check_core: sa is null\n");
			result = IPSEC_ACTION_DROP;
		}
	}

	/* If renew_sa_list is up, policy's sa_index_list is renewed by sa_find_by_sa_index. */
	if (renew_sa_list) {
		error = spd_find_same_spd(policy, SPD_WITH_WLOCK);
		if (error == -EEXIST) {
			sa_list_clear(&policy->sa_index_list);
			sa_list_copy(&policy->sa_index_list, sa_index_list);
			SPD_WUNLOCK(&policy->lock);
		}
	}

	*ipsec_ptr = sa_index_list;
	ipsec6_pr("ipsec6_output_check_core: end\n");	

err:
	return result;
}

int ipsec6_output_check(struct sock *sk, struct flowi *fl, void **ipsec_ptr)
{
	struct in6_addr *saddr,*daddr;
	u16 sport,dport;
	unsigned char proto;
	struct selector selector;
	int result = IPSEC_ACTION_BYPASS; 	/* default */

	ipsec6_pr("ipsec6_output_check: called\n");
	if (!sk && !fl) {
		printk(KERN_ERR "flowi and sock are NULL\n");
		result = -EINVAL;
		goto err;
	}
	
	if (fl && fl->fl6_src) {
		saddr = fl->fl6_src; 
	} else {
		if (sk) {
			saddr = &sk->net_pinfo.af_inet6.saddr; 
		} else {
			result = -EINVAL;
			printk(KERN_ERR "sock is null\n");
			goto err;
		}
	}

	if (fl && fl->fl6_dst) {
		daddr = fl->fl6_dst; 
	} else {
		if (sk) {
			daddr = &sk->net_pinfo.af_inet6.daddr; 
		} else { 
			result = -EINVAL;
			printk(KERN_ERR "flowi and sock are NULL\n");
			goto err;
		}
	}

	if (fl) { 
		sport=fl->uli_u.ports.sport;
		dport=fl->uli_u.ports.dport;
		proto=fl->proto;
	} else if (sk) {
		sport=sk->sport;
		dport=sk->dport;
		proto=sk->protocol;
	} else {
		result = -EINVAL;
		printk(KERN_ERR "flowi and sock are NULL\n");
		goto err;
	}

	/* XXX have to decide to the policy of ICMP messages -mk*/
	if (proto == IPPROTO_ICMPV6) {
		sport = 0;
		dport = 0;
#ifndef CONFIG_APPLY_ICMPV6_IPSEC /* XXX currently we can't handle NS-NA -mk */
		ipsec6_pr("ipsec6_output_check: skip ICMP packet!\n");
		goto err; 
#endif /* CONFIG_APPLY_ICMPV6_IPSEC */
	}

	/* XXX config  port policy */
	memset(&selector, 0, sizeof(struct selector));

	((struct sockaddr_in6 *)&(selector.src))->sin6_family = AF_INET6;
	ipv6_addr_copy(&((struct sockaddr_in6 *)&(selector.src))->sin6_addr,
		       saddr);
	((struct sockaddr_in6 *)&(selector.dst))->sin6_family = AF_INET6;
	ipv6_addr_copy(&((struct sockaddr_in6 *)&(selector.dst))->sin6_addr,
		       daddr);
	selector.proto = proto;
	selector.prefixlen_d = 128;
	selector.prefixlen_s = 128;

	((struct sockaddr_in6 *)&(selector.src))->sin6_port = sport;	
	((struct sockaddr_in6 *)&(selector.dst))->sin6_port = dport;	

#ifdef CONFIG_IPV6_IPSEC_DEBUG
	{
		char buf[64];
		ipsec6_pr("ipsec6_output_check: original dst addr: %s\n", in6_ntop(daddr, buf));
		ipsec6_pr("ipsec6_output_check: original dst port: %u\n", ntohs(dport));
		ipsec6_pr("ipsec6_output_check: original src addr: %s\n", in6_ntop(saddr, buf));
		ipsec6_pr("ipsec6_output_check: original src port: %u\n", ntohs(sport));

		ipsec6_pr("ipsec6_output_check: selector dst addr: %s\n", 
				in6_ntop( &((struct sockaddr_in6 *)&(selector.dst))->sin6_addr, buf));
		ipsec6_pr("ipsec6_output_check: selector dst port: %u\n", 
				ntohs(((struct sockaddr_in6 *)&(selector.dst))->sin6_port));
		ipsec6_pr("ipsec6_output_check: selector src addr: %s\n", 
				in6_ntop( &((struct sockaddr_in6 *)&(selector.src))->sin6_addr, buf));
		ipsec6_pr("ipsec6_output_check: selector src port: %u\n", 
				ntohs(((struct sockaddr_in6 *)&(selector.src))->sin6_port));
		ipsec6_pr("ipsec6_output_check: selector proto: %u\n", selector.proto);
	}
#endif /* CONFIG_IPV6_IPSEC_DEBUG */

	result = ipsec6_output_check_core(&selector, ipsec_ptr);

 err:
		return result;
}



/* XXX XXX XXX  -mk */
int ipsec6_forward_check(struct sk_buff* skb) 
{
        u16 sport=0, dport=0;
	int rtn = 0;
        struct ip6_packet_refs refs; /* XXX */
	struct in6_addr *saddr, *daddr;
//	struct ipsec_sa *ipsec_sa = NULL;

        if (!skb) {
		printk(KERN_ERR "ipsec6_forward_check: skb is null.\n");
		rtn = -EINVAL;
	}

        if ( (skb->security & RCV_AUTH) || (skb->security & RCV_CRYPT) ) {
		printk(KERN_ERR "ipsec6_forward_check: already processed by IPSec\n");
                rtn =  -EINVAL;
        }

        analyze_ip6_packet((char*)skb->nh.ipv6h, &refs);

	daddr = (struct in6_addr *)&(refs.ipv6hdr->daddr);
	saddr = (struct in6_addr *)&(refs.ipv6hdr->saddr);

        if (refs.resttype == NEXTHDR_TCP) {
                struct tcphdr *th = (struct tcphdr*)refs.rest;

                sport = th->source;
                dport = th->dest;
                ipsec6_pr("TCP packet: sport=%d, dport=%d\n", ntohs(sport),ntohs(dport));
        } else if (refs.resttype == NEXTHDR_UDP) {
                struct udphdr *uh = (struct udphdr*)refs.rest;

                sport = uh->source;
                dport = uh->dest;
                ipsec6_pr("UDP packet: sport=%d, dport=%d\n", ntohs(sport),ntohs(dport));
        }

	/* AH */
	/* TODO: lookup spd using by dst addr, dst port, src addr, src port */
	
	/* ESP */
	/* TODO: lookup spd using by dst addr, dst port, src addr, src port */

        return rtn;
}

